
/*
 * Copyright 1992, 1993 by Kaleb S. Keithley
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Kaleb S. Keithley makes no 
 * representations about the suitability of this software for 
 * any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

#include <sys/ioctl.h>
#ifdef SVR4
#include <sys/fbio.h>
#else
#include <sun/fbio.h>
#endif
#include <fcntl.h>
#include <stdio.h>
#include "sun.h"

char *fb_type_names[FBTYPE_LASTPLUSONE] = {
    "FBTYPE_SUN1BW (bwone) not supported",
    "FBTYPE_SUN1COLOR (cgone1) not supported",
    "FBTYPE_SUN2BW (bwtwo)",
    "FBTYPE_SUN2COLOR (cgtwo)",
    "FBTYPE_SUN2GP (gpone/gptwo) not supported",
    "FBTYPE_SUN5COLOR (386i accelerated) not supported",
    "FBTYPE_SUN3COLOR (cgthree)",
    "FBTYPE_MEMCOLOR -- not supported",
    "FBTYPE_SUN4COLOR (cgfour)",
    "FBTYPE_NOTSUN1 -- not supported",
    "FBTYPE_NOTSUN2 -- not supported",
    "FBTYPE_NOTSUN3 -- not supported",
#ifdef FBTYPE_SUNFAST_COLOR
    "FBTYPE_SUNFAST_COLOR (cgsix/GX)",
#endif
#ifdef FBTYPE_SUNROP_COLOR
    "FBTYPE_SUNROP_COLOR (cgeight/cgnine) -- not supported",
#endif
#ifdef FBTYPE_SUNFB_VIDEO
    "FBTYPE_SUNFB_VIDEO -- not supported",
#endif
#ifdef FBTYPE_SUNGIFB
    "FBTYPE_SUNGIFB -- not supported",
#endif
#ifdef FBTYPE_SUNGPLAS
    "FBTYPE_SUNGPLAS -- not supported",
#endif
#ifdef FBTYPE_SUNGP3
    "FBTYPE_SUNGP3 (cgtwelve/GS) not supported",
#endif
#ifdef FBTYPE_SUNGT
    "FBTYPE_SUNGT (gt) -- not supported",
#endif
#ifdef FBTYPE_RESERVED2
    "FBTYPE_RESERVED2 not supported",
#endif
#ifdef FBTYPE_RESERVED1
    "FBTYPE_RESERVED1 not supported"
#endif
};

char *device_names[] = {
#ifndef i386
    CGTWO0DEV, CGTWO1DEV, CGTWO2DEV,
#endif
    CGTHREE0DEV,
#ifndef i386
    CGTHREE1DEV, CGTHREE2DEV,
#ifdef FBTYPE_SUNFAST_COLOR
    CGSIX0DEV, CGSIX1DEV, CGSIX2DEV,
#endif
    CGFOUR0DEV, BWTWO0DEV, BWTWO1DEV
#endif  /* i386 */
};

#define MAX_DEV_NAMES sizeof device_names / sizeof device_names[0]

static int getFbType (
#ifdef __STDC__
char *, int *
#endif
);

#ifdef __STDC__
int main (
    int		argc,
    char	**argv)
#else
int main (argc, argv)
    int		argc;
    char	**argv;
#endif

{
    int		fb_type[5], loop;

    if (getFbType ("/dev/fb", fb_type) == -1)
	(void) printf ("%s\n", "error opening '/dev/fb'");
    else {
	if (fb_type[0] >= 0 && fb_type[0] < FBTYPE_LASTPLUSONE) {
	    int		index = 1;
	    (void) printf ("\n\n'/dev/fb' is a %s\n", fb_type_names[fb_type[0]]);
	    for (loop = 0; loop < FB_ATTR_NEMUTYPES; loop++, index++)
		if (fb_type[index] >= 0 && fb_type[index] < FBTYPE_LASTPLUSONE)
		    (void) printf ("  and it may emulate %s\n", fb_type_names[fb_type[index]]);
    	} else
	    (void) printf ("\n\n'/dev/fb' is an unknown frame buffer type\n");
    }
    (void) printf ("\n%s\n\n", 
"The following supported devices were found, which Xsun will\n\
attempt to use unless over-ridden with the -dev command line\n\
option or the XDEVICE environment variable:");
    for (loop = 0; loop < MAX_DEV_NAMES; loop++) {
	if (getFbType (device_names[loop], fb_type) != -1) {
	    if (fb_type[0] >= 0 && fb_type[0] < FBTYPE_LASTPLUSONE)
		(void) printf ("  '%s' is a %s\n", device_names[loop],
				fb_type_names[fb_type[0]]);
	    else
		(void) printf ("  '%s' is an unknown frame buffer type\n");
	}
    }
    return 0;
}

#ifdef __STDC__
static int getFbType (
    char	*fbname,
    int		*fbtype)
#else
static int getFbType (fbname, fbtype)
    char	*fbname;
    int		*fbtype;
#endif
{
    struct fbgattr fbattr;
    int		fd, ioctl_ret, ret = -1;

    if ((fd = open (fbname, O_RDWR, 0)) != -1) {
	/* FBIOGATTR fails for early frame buffer types */
	if ((ioctl_ret = ioctl(fd,FBIOGATTR,&fbattr)) == -1)
	    ioctl_ret = ioctl(fd, FBIOGTYPE, &fbattr.fbtype);
	if (ioctl_ret == -1)
	    *fbtype = -1;
	else {
	    fbtype[0] = fbattr.fbtype.fb_type;
	    fbtype[1] = fbattr.emu_types[0];
	    fbtype[2] = fbattr.emu_types[1];
	    fbtype[3] = fbattr.emu_types[2];
	    fbtype[4] = fbattr.emu_types[3];
	}
	close(fd);
	ret = 0;
    }
    return ret;
}
